from clumper import Clumper
from parse import parse
from pathlib import Path


def trim_nodeid(nodeid):
    """
    Trims a nodeid from a pytest run such that all parametrized tests fold into a single id. 
    
    Usage
    
    ```python
    nodeid = "test_shape_trained_model[random_xy_dataset_regr15]"
    expected = "test_shape_trained_model[*]"
    assert trim_nodeid(nodeid) == expected
    ```
    """
    if "[" not in nodeid:
        return nodeid
    return nodeid[:nodeid.find("[")] + "[*]"


def parse_test_info(clump, trim=True):
    """
    Parses the test info for useful information.
    """
    c = clump.keep(lambda d: "duration" in d.keys())
    if trim:
         c = c.mutate(nodeid=lambda d: trim_nodeid(d['nodeid']))
    return (c
            .group_by("nodeid")
            .agg(duration=("duration", "sum"))
            .mutate(duration=lambda d: round(d['duration'] * 1000),
                    parsed=lambda d: parse("{path}/{file}.py::{test}", d['nodeid']).named,
                    parsed_path=lambda d: f"{d['parsed']['path']}/{d['parsed']['file']}.py",
                    parsed_test=lambda d: d['parsed']['test'],
                    path=lambda d: d['parsed']['path'],
                    hierarchy=lambda d: tuple(list(Path(d['parsed_path']).parts) + [d['parsed_test']]))
            .drop("parsed"))


def to_hierarchy_dict(clump, hierarchy_col=None, value_col=None, value_name="value"):
    """
    Turns a clumper into a hierarchical dictionary ready for d3.
    """
    lines = clump.ungroup().sort(lambda d: tuple(d[hierarchy_col])).select(hierarchy_col, value_col).collect()
    data = {"name": "root", "children": []}
    for line in lines: 
        cursor = data
        for idx, item in enumerate(line['hierarchy']):
            res = [c for c in cursor["children"] if c['name'] == item]
            if len(res) == 0:
                if idx != len(line['hierarchy']) - 1:
                    cursor['children'].append({"name": item, "children": []})
                else:
                    cursor['children'].append({"name": item, value_name: line['duration']})
                res = [c for c in cursor["children"] if c['name'] == item]
            cursor = res[0]
    return data


class Node:
    def __init__(self, name, children=None, val=None):
        self.name = name
        self.children = children
        self.val = val
    
    @property
    def value(self):
        if self.val:
            return self.val
        return sum([c.value for c in self.children])
    
    @classmethod
    def from_dict(cls, d):
        if 'children' in d:
            return Node(name=d['name'], 
                        children=[cls.from_dict(c) for c in d['children']])
        return Node(name=d['name'], val=d['value'])
    
    def to_value_dict(self):
        if self.children:
            return {"name":self.name, 
                    "value": self.value, 
                    "children": sorted([c.to_value_dict() for c in self.children], key=lambda d: -d['value'])}
        return {"name":self.name, "value": self.value}



import os 
from setuptools import setup, find_packages

from pytest_duration_insights import VERSION

base_packages = ["pytest-reportlog>=0.1.2", "clumper>=0.2.12", "parse>=1.19.0", "typer>=0.3.2"]


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="pytest-duration-insights",
    version=VERSION,
    packages=find_packages(exclude=['notebooks']),
    install_requires=base_packages,
    entry_points={
        'console_scripts': ['pytest-duration-insights=pytest_duration_insights.__main__:app']
    },
    package_data={"pytest-duration-insights": ["static/**/*"]},
    long_description=read("readme.md"),
    project_urls={
        'Bug Tracker': 'https://github.com/koaning/pytest-duration-insights/issues',
        'Documentation': 'https://github.com/koaning/pytest-duration-insights',
        'Source Code': 'https://github.com/koaning/pytest-duration-insights'
    },
)



import sys
import typer
import shutil
import tempfile
import subprocess
from pathlib import Path
from pkg_resources import resource_filename

from clumper import Clumper 

from ._parsing import to_hierarchy_dict, parse_test_info, Node
from pytest_duration_insights import VERSION


app = typer.Typer(
    name="pytest duration insights", 
    add_completion=False, 
    help="""
    This app allows you to inspect and explore pytest run times.

    Assuming you've got `pytest-reportlog` installed you can collect
    test run data and visualise it via;

    pytest --report-log reportlog.jsonl

    pytest-duration-insights explore reportlog.jsonl
    """
)

@app.command()
def explore(report_path: str = typer.Argument(..., help="Report log to visualise."), 
            no_trim: bool = typer.Option(False, is_flag=True, help="Flag to not reduce parametrized calls."), 
            port: int = typer.Option(8002, help="Port to serve the dashboard on.")):
    """Starts up a pytest reportlog dashboard."""
    # Use clumper to arrange the data appropriately.
    res = (Clumper.read_jsonl(report_path)
              .pipe(parse_test_info, trim=not no_trim)
              .pipe(to_hierarchy_dict, hierarchy_col="hierarchy", value_col="duration"))
    
    # We server everything as static files from a temporary folder
    tmpdir = tempfile.mkdtemp()
    shutil.copytree(src=Path(__file__).parent / "static", 
                    dst=Path(tmpdir) / "static")
    Clumper(res, listify=False).write_json(Path(tmpdir) / "static" / "data.json")
    tree_res = Node.from_dict(res).to_value_dict()
    Clumper(tree_res, listify=False).write_json(Path(tmpdir) / "static" / "treedata.json")
    
    # This a bit hacky but does the job
    subprocess.run(["python", "-m", "http.server", str(port), "--directory", str(Path(tmpdir) / "static")])

@app.command()
def version():
    """Returns the version."""
    return VERSION
       

if __name__ == "__main__":
    """
    pytest-duration-insights --report-file --open-browser --reduce
    """
    app()
